-- Time-stamp: <Mon Dec 12 2005 23:24:04 hwloidl>
--
-- Interpreter for the Intermediate Language of the MiniJava compiler
---------------------------------------------------------------------------

module Interp where

import AbsSyn
--import Parse
import Data.Map as Map
import Control.Monad

import Debug.Trace(trace)

type RegMap = Map Register Val
type Frame = Integer
type Offset = Integer
type Memory = Map Address Val
type Address = Integer
type Trace = String
data Val = I !Integer |
           A !Address |
           L !Label
     deriving (Read,Show,Eq)
type Msgs = [String]

data TraceOpts = CallTrace | JumpTrace | MoveTrace | AllocTrace | RegMaps | MemMaps | LabMaps | DebugArgs
 deriving (Read,Show,Eq)
type Settings = ([TraceOpts])

type LabTab = Map Label [Stm]
type FunTab = Map Label [Stm]
type FunParams = Map Label [Access]


type State = (RegMap,Memory,LabTab,FunTab,FunParams,Msgs,Settings)

-- evaluation monad
data Intp a = T (State -> (State, a))

runIntp :: State -> Intp a -> a
runIntp s (T f) = snd (f s)

instance Monad Intp where
 return x = T (\ s -> (s, x))
 (T t) >>= f =  T (\ s -> let { (s1,x1) = t s ; T f' = f x1 } in f' s1)

instance Num Val where
 (A x) + (A y) = A (x+y)
 (A x) + (I y) = A (x+y)
 (I x) + (A y) = A (x+y)
 (I x) + (I y) = I (x+y)
 (A x) * (A y) = A (x*y)
 (A x) * (I y) = A (x*y)
 (I x) * (A y) = A (x*y)
 (I x) * (I y) = I (x*y)
 fromInteger n = I n

instance Ord Val where
 (A x) <= (A y) = x<=y
 (I x) <= (I y) = x<=y
 _ <= _ = error "Cannot compare address with integer"
 
emptyEnv :: RegMap
emptyEnv = empty

emptyTrace :: Msgs
emptyTrace = []

defaultSettings :: Settings
defaultSettings = ([CallTrace,MoveTrace,AllocTrace])

emptyState :: State
emptyState = (Map.empty, Map.empty, 
              Map.empty, Map.empty, Map.empty,
              emptyTrace, defaultSettings)

combineStates (rm, mm, lT, fT, pT, ts, ss) (rm', mm', lT', fT', pT', ts', ss') =
  (Map.union rm rm', Map.union mm mm', Map.union lT lT', Map.union fT fT', Map.union pT pT', ts++ts', ss++ss')

-- type State = (RegMap,Memory,LabTab,FunTab,FunParams,Msgs,Settings)
initState :: [AbsSyn.Fct] -> State
initState = (foldr combineStates emptyState) .
            (Prelude.map ( \ (Fct name ps ss) -> (Map.empty, Map.empty, -- empty regs and heap
	                      runLabTabFiller (fillTab ss),
                              Map.union Map.empty (Map.singleton name ss), -- body
                              Map.union Map.empty (Map.singleton name (Prelude.map param2access ps)), -- params
                              emptyTrace, defaultSettings) ))

-- return value of functions
retReg :: String 
retReg = "t2"

retRegA :: Access
retRegA = InReg retReg


thisReg :: String 
thisReg = "t1"

thisRegA :: Access
thisRegA = InReg thisReg

fpReg :: String 
fpReg = "t32"

spReg :: String 
spReg = "t98" -- ???

hpReg :: String 
hpReg = "t99" -- ???

-- dummy vals; mainly to catch errors of uninit vars
dummyInt = (I 96)
dummyAddr = (A 0)

dummyRegVal = (I 81)
dummyMemVal = (I 82)

spStart = (A 8000)
hpStart = (A 9000)
spEnd = (A 1024)
hpEnd = (A 65535)

---

showRegMap :: Intp ()
showRegMap = getR >>= \ rm -> msgs ["Register Mapping: ", (showTree rm)]

showMemMap :: Intp ()
showMemMap = getM >>= \ rm -> msgs ["Memory Mapping: ", (showTree rm)]

showLabTab :: Intp ()
showLabTab = labTab >>= \ lT -> msgs ["labTab: ", (showTree lT)]

showFunTab :: Intp ()
showFunTab = funTab >>= \ fT -> msgs ["funTab: ", (showTree fT)]

execCall :: [AbsSyn.Fct] -> AbsSyn.Fct -> [Integer] -> ([String],Val)
execCall prg (Fct name ps ss) es = let
                                     state = initState prg
                                     code = do
                                              showLabTab
                                              showFunTab
	                                      rm <- initRegs
                                              setR rm
	                                      rm <- getR 
                                              -- trace ("Regs after init: " ++ (showTree rm)) (return ())
                                              --trace ("Executing: " ++ (show (CALL (NAME name) ((TEMP thisReg):(Prelude.map (\ n -> CONST n) es))))) (return ())
                                              res <- eval (CALL (NAME name) ((TEMP thisReg):(Prelude.map (\ n -> CONST n) es)))
	                                      t <- getT
                                              return (reverse t,res)

				   in
                                     runIntp state code
	                             

dummyVal :: Val
dummyVal = I 99

msg :: String -> Intp ()
msg str = T $ \ (e,m,lT,fT,fP,t,s) -> ((e,m,lT,fT,fP,str:t,s), ())

msgs :: [String] -> Intp ()
msgs xs = sequence_ (Prelude.map msg xs)

getR = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,e)
getM = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,m)
getT = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,t)
getS = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,s)
labTab = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,lT)
funTab = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,fT)
funParams = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st,fP)

setR x = T $ \ st@(e,m,lT,fT,fP,t,s) -> ((x,m,lT,fT,fP,t,s),())
setM x = T $ \ st@(e,m,lT,fT,fP,t,s) -> ((e,x,lT,fT,fP,t,s),())
setT x = T $ \ st@(e,m,lT,fT,fP,t,s) -> ((e,m,lT,fT,fP,x,s),())

getReg :: Register -> Intp Val
getReg r = getR >>= \ e -> let x = Map.findWithDefault dummyRegVal r e in return x

getRegSafe :: Register -> Intp Val
getRegSafe a = getReg a >>= \ x -> if x==dummyMemVal then error ("PANIC: lookup to uninitialised register " ++ (show a)) else return x

setReg :: Register -> Val -> Intp ()
setReg r x = getR >>= \ rm -> setR (Map.union (Map.singleton r x) rm)

getMem :: Address -> Intp Val
getMem a = getM >>= \ m -> let x = Map.findWithDefault dummyMemVal a m in  return x
--getMem a = T $ \ st@(e,m,lT,fT,fP,t,s) -> (st, Map.findWithDefault 81 a m)

getMemSafe :: Address -> Intp Val
getMemSafe a = getMem a >>= \ x -> if x==dummyMemVal then error ("PANIC: lookup to uninitialised mem cell " ++ (show a)) else return x

getLabCode :: Label -> Intp [Stm]
getLabCode l = labTab >>= \ lT -> let x = Map.findWithDefault [] l lT in  return x

getFunCode :: String -> Intp [Stm]
getFunCode f = funTab >>= \ fT -> let x = Map.findWithDefault [] f fT in  return x

getFunParams :: String -> Intp [Access]
getFunParams f = funParams >>= \ fP -> let x = Map.findWithDefault [] f fP in  return x

getTraceOpts :: Intp [TraceOpts]
getTraceOpts = getS >>= \ (topts) -> return topts

type BState = (String, [Stm], Map String [Stm])
newtype LabTabFiller a = B (BState -> (BState, a))

addBB :: Stm -> LabTabFiller ()
addBB bb = B $ \ (currLab, currBB, m) -> ((currLab, currBB++[bb], m),())

getCurrBB = B $ \ (currLab, currBB, m) -> ((currLab, currBB, m), currBB)

runLabTabFiller :: LabTabFiller a -> Map String [Stm]
runLabTabFiller (B f) = let ((_,_,m),_) = f ("", [], Map.empty) in m

instance Monad LabTabFiller where
 return x = B (\ s -> (s, x))
 (B t) >>= f =  B (\ s -> let { (s1,x1) = t s ; B f' = f x1 } in f' s1)


class Evaluator x where
 eval :: x -> Intp Val
 fillTab :: x -> LabTabFiller ()
 {-
 prints :: [x] -> IO ()
 prints [] = return ()
 prints (e1:es) = runIntp emptyState (eval e1 >>= \ (_,t1,_) -> eval es >>= \ (_,t2,_) -> return (t1++"\n"++t2++"\n")) >>= \ str -> putStrLn str
 -}

instance Evaluator Stm where
 eval c@(MOVE dest src) = do 
                           -- trace ("Entering " ++ (show c) ++"\n") return ()
                           v <- eval src
                           access <- lval dest
                           store access v
                           topts <- getTraceOpts 
                           let traceme = MoveTrace `elem` topts
                           let rmaps = RegMaps `elem` topts
                           let mmaps = MemMaps `elem` topts
                           if traceme
                             then 
                               case (access) of {
                                (InReg _) -> getR >>= \ regs -> (if rmaps then trace (".. Registers after " ++ (show c) ++ showTree regs ++"\n") else id) return () ;
                                (InMem _) -> getM >>= \ regs -> (if mmaps then trace (".. Memory after " ++ (show c) ++ showTree regs ++"\n") else id) return () }
                             else return ()
                           return dummyVal

 eval (EXP e) = eval e

 eval (JUMP dest poss) = do  -- ToDo: test poss
                          (L lab) <- eval dest
                          stm <- getLabCode lab
                          topts <- getTraceOpts 
                          let traceme = JumpTrace `elem` topts
                          let maps = LabMaps `elem` topts
                          if traceme 
                            then trace ("-> JUMP to label " ++ lab ++ (if maps then " continues with code " ++ (show stm) else "") ++"\n") return ()
                            else return ()
                          eval stm

 eval (CJUMP rel left right trueLab falseLab) =
   do
     (I x) <- eval left
     (I y) <- eval right
     let z = case rel of
              AbsSyn.EQ -> x==y
              AbsSyn.NE -> x/=y
              AbsSyn.LT -> x<y
              AbsSyn.GT -> x>y
              AbsSyn.LE -> x<=y
              AbsSyn.GE -> x>=y
              AbsSyn.ULT -> x<y
              AbsSyn.ULE -> x<=y
              AbsSyn.UGT -> x>y
              AbsSyn.UGE -> x>=y
     
     topts <- getTraceOpts 
     let traceme = JumpTrace `elem` topts
     let maps = LabMaps `elem` topts
     if traceme 
       then trace ("__ CJUMP results in " ++ (show x ) ++ " " ++ (show rel) ++ " " ++ (show y) ++ " which is " ++ (show z)++"\t jumping to label " ++ (if z then trueLab else falseLab)) return ()
       else  return ()
     if z
       then getLabCode trueLab >>= \ c -> eval c
       else getLabCode falseLab >>= \ c -> eval c

 eval (SEQ first second) = do 
                             eval first
                             eval second

 eval (LABEL l) =  return (L l)

 fillTab this@(MOVE dest src) = addBB this
 fillTab this@(EXP e) = addBB this
 fillTab this@(JUMP dest poss) = addBB this
 fillTab this@(CJUMP rel left right trueLab falseLab) = addBB this
 fillTab this@(SEQ first second) = addBB this
 fillTab this@(LABEL l) = B $ \ (currLab, currBB, labTab) ->
                                   if (currLab/="")
                                     then ((l,[],Map.union (Map.singleton currLab currBB) labTab),()) 
                                     else ((l,[],Map.empty), ())


intOp :: AbsSyn.BinOp -> Integer -> Integer -> Integer
intOp AbsSyn.PLUS x y  = x+y
intOp AbsSyn.MINUS x y  = x-y
intOp AbsSyn.MUL x y  = x*y
intOp AbsSyn.DIV x y  = x `div` y
intOp AbsSyn.AND x y  = boolToVal ((valToBool x)&&(valToBool y))
intOp AbsSyn.OR x y  = boolToVal ((valToBool x)||(valToBool y))
intOp AbsSyn.LSHIFT x y  = x*(mypow 2 y)
intOp AbsSyn.RSHIFT x y  = x `div` (mypow 2 y)
intOp AbsSyn.XOR x y  = boolToVal (myxor (valToBool x) (valToBool y))
intOp op x y = error ((show op) ++" not yet implemented")

instance Evaluator Exp where
 --eval :: Env -> Expr -> (Val, Env)
 eval (CONST value) = return (I value)
 eval (NAME label) = return (L label)
 eval (TEMP temp) = do
                     x <- getRegSafe temp
                     return x
 eval (BINOP op left right)=do
                             lhs <- eval left
                             rhs <- eval right
                             return $
                               case lhs of
                                 (I x) -> case rhs of
                                           (I y) -> (I (intOp op x y))
                                           (A a) -> (A (intOp op x a))
                                 (A a) -> case rhs of
                                           (I y) -> (A (intOp op a y))
                                           (A b) -> (A (intOp op a b))
 eval (MEM addr) = do 
                         (A a) <- eval addr
                         z <- getMemSafe a
                         return z

 eval c@(CALL func args) = do
                            -- trace ("Entering " ++ (show c) ++ "\n") return ()
                            (L f) <- eval func
                            q <- catchExternal f args
                            case q of 
                              (Just a) -> return a
                              Nothing -> do
                            	fc <- getFunCode f
                            	ps <- getFunParams f
                            	{-
                            	rm <- getR
                            	rm' <- initRegs
                            	setR (Map.union rm rm')
                            	-}
                            	fp <- getReg fpReg
                            	sp <- getReg spReg
                            	hp <- getReg hpReg
                                old_regs <- getR -- NB: all registers are callee saved
                                -- old_mem <- getM -- NO: side effects on heap possb
                            	-- trace (".. Allocating frame; current frame at " ++ (show fp) ++"\n") return ()
                                unless ((length ps)==(length args)) $
                                  trace ("** call " ++ (show f) ++ " mismatching arg lengths: " ++ (show ps) ++ (show args)) (return ())
                            	vals <- passArgs ps args -- vals for tracing only
                            	r <- getR
                            	topts <- getTraceOpts 
                            	let traceme = CallTrace `elem` topts
                            	let maps = RegMaps `elem` topts
                            	when traceme $
                            	  trace ("=> CALL " ++ (show f) ++ (show vals) ++ "\tFP= " ++ (show fp) ++ "\tSP= " ++ (show sp) ++ "\tHP= " ++ (show hp) ++ "\n" ++ (if maps then "RegMap after passing args: " ++ (show args) ++ " -> " ++ (show ps) ++"\n" ++ (showTree r) else "")) return ()
                            	allocFrame 100
                                {-
                            	if traceme 
                            	  then getReg fpReg >>= \ fp -> getReg spReg >>= \ sp -> trace ("** new FP= " ++ (show fp) ++"\tnew SP= " ++ (show sp) ++ "\n") return ()
                                  else return ()
                                -}
                            	eval fc 
                            	r <- getRegSafe retReg
                            	deallocFrame
                                {-
                            	if traceme 
                            	  then getReg fpReg >>= \ fp -> getReg spReg >>= \ sp -> trace ("** old FP= " ++ (show fp) ++"\told SP= " ++ (show sp) ++ "\n") return ()
                                  else return ()
                                -}
                            	when traceme $
                            	  trace ("=> CALL " ++ (show f) ++ (show vals) ++ " returns value " ++ (show r) ++ "\n") (return ())
                                -- setM old_mem
                                setR old_regs
                            	return r

                         {- let
                           argvals = scanl (foldl (\ (v,q,t) x. eval t q x) (dummyVal,q,t) args
                           (q',t') = bindArgs (thd3 (last argvals)) (snd3 (last argvals)) argvals
                         in
                           eval t' q' (funTab func) -}

 eval (ESEQ stm exp) = eval stm >> eval exp

 fillTab (CONST value) = return ()
 fillTab (NAME label) = return ()
 fillTab (TEMP temp) = return ()
 fillTab (BINOP op left right)= return ()
 fillTab (MEM addr) =  return ()
 fillTab (CALL func args) =  return ()
 fillTab (ESEQ stm exp) =  return ()

instance Evaluator Fct where
 eval (Fct l ps ss) = error "Can't eval function declaration"
 fillTab (Fct l ps ss) = fillTab ss

instance Evaluator [Stm] where
 eval [] = return dummyVal
 eval (x@(JUMP _ _):xs) = eval x
 eval (x@(CJUMP _ _ _ _ _):xs) = eval x
 eval (x:xs) = eval x >> eval xs
 fillTab [] = B $ \ (currLab, currBB, labTab) ->
                                   if (currLab/="")
                                     then (("",[],Map.union (Map.singleton currLab currBB) labTab),()) 
                                     else (("",[],Map.empty), ())
 fillTab (x:xs) = fillTab x >> fillTab xs

instance  Evaluator [Exp] where
 eval [] = return dummyVal
 eval (x:xs) = eval x >> eval xs
 fillTab [] = B $ \ (currLab, currBB, labTab) ->
                                   if (currLab/="")
                                     then (("",[],Map.union (Map.singleton currLab currBB) labTab),()) 
                                     else (("",[],Map.empty), ())
 fillTab (x:xs) = fillTab x >> fillTab xs
 
---------------------------------------------------------------------------
-- Mem mgmt:

data Access = InReg Register
            | InFrame (Integer, Integer) -- (Frame, Offset)
            | InMem Address
 deriving (Read,Show,Eq)

lval :: Exp -> Intp Access
lval (TEMP r) = return (InReg r)
lval (MEM e) = eval e >>= \ a -> case a of { (A a) -> return (InMem a) ; (L l) -> error ("PANIC: lval returned label " ++ l) ; ; (I l) -> error ("PANIC: lval returned integer " ++ (show l)) }
lval e = error ("Illegal lval in " ++ (show e))

store :: Access -> Val -> Intp Val
store (InReg r) x = getR >>= \ rm -> setR (Map.union (Map.singleton r x) rm) >>  return (dummyVal)
store (InFrame (0,o)) x = getM >>= \ mm -> getReg fpReg >>= \ (A fpa) -> setM (Map.union (Map.singleton (fpa+o) x) mm) >>  return (dummyVal)
store (InFrame (1,o)) x = getM >>= \ mm -> getReg spReg >>= \ (A spa) -> setM (Map.union (Map.singleton (spa+o) x) mm) >>  return (dummyVal)
store (InMem a) x = getM >>= \ mm -> getReg fpReg >>= \ fp -> setM (Map.union (Map.singleton a x) mm) >>  return (dummyVal)
store a x = error $  "PANIC: store: illegal access of destination: "++(show a)

storeA (A a) x = T $ \ (e,m,lT,fT,fP,t,s) -> ((e,Map.insert a x m,lT,fT,fP,t,s), dummyVal)
storeA _ x = error "Cannot store into a non-address"

param2access :: AbsSyn.Param -> Access
param2access (LOC n) = InFrame (1,n) -- 1 means one frame off; i.e. take sp as basis in store!!
param2access (REG t) = InReg t
--param2access (ADR a) = InMem t

passArgs :: [Access] -> [Exp] -> Intp [Val]
passArgs [] [] = return []
passArgs (p:ps) (e:es) = do 
	                   x <- eval e
                           store p x
	                   xs <- passArgs ps es
                           topts <- getTraceOpts 
                           let traceme = DebugArgs `elem` topts
                           when traceme $
                             trace ("passArgs: " ++ (show x) ++ " -> " ++ (show p)) (return ())
                           return (x:xs)
passArgs _ _ = error "PANIC: passArgs: Mismatch of lengths of ars and param lists"

-- names of callee save registers
calleeSaves :: [String]
calleeSaves = ["t40"]

allocFrame :: Integer ->  Intp ()
allocFrame sz = do
              	 sp@(A spa) <- getReg spReg
              	 fp@(A fpa) <- getReg fpReg
	         when (sp+(I sz)<spEnd) $
                   error "PANIC: stack overflow"
                 store (InMem spa) fp -- (A fpa)
	      	 setReg fpReg (A spa)
	      	 setReg spReg (A (spa-sz))

deallocFrame :: Intp ()
deallocFrame = do
              	(A spa) <- getReg spReg
                (A fpa) <- getReg fpReg
	        setReg spReg (A fpa)
	        m <- getMem fpa
	        setReg fpReg m

initRegs :: Intp (Map String Val)
initRegs = do 
             let xs = Prelude.map ( \ x -> (x, dummyVal) ) calleeSaves 
	     return $ Map.fromList ([(thisReg, dummyAddr),(retReg, dummyInt), (fpReg, spStart), (spReg, spStart), (hpReg, hpStart)]++xs)
             -- return (foldr Map.union Map.empty xs') 

---------------------------------------------------------------------------
-- handling of external calls

catchExternal :: String -> [Exp] -> Intp (Maybe Val)
catchExternal "L_halloc_obj" es = do
                                  vals <- mapM eval es 
                                  a <- case vals of
                                    	 [(I n)] -> halloc_obj n
                                    	 _ -> error ("PANIC: illegal arguments to _halloc_obj (should be 1 integer argument): " ++ (show vals))
                                  return (Just a)
catchExternal "L_init_obj" es = do
                                  vals <- mapM eval es 
                                  case vals of
                                    [(A a), (I n)] -> init_obj a n
                                    _ -> error ("PANIC: illegal arguments to _init_obj (should be 1 address and 1 integer argument): " ++ (show vals))
                                  return (Just dummyAddr)
catchExternal "L_halloc_arr" es = do
                                  vals <- mapM eval es 
                                  a <- case vals of
                                    	 [(I n)] -> halloc_arr n
                                    	 _ -> error ("PANIC: illegal arguments to _halloc_arr (should be 1 integer argument): " ++ (show vals))
                                  return (Just a)
catchExternal "L_init_arr" es = do
                                  vals <- mapM eval es 
                                  case vals of
                                    [(A a), (I n)] -> init_arr a n
                                    _ -> error ("PANIC: illegal arguments to _init_arr (should be 1 address and 1 integer argument): " ++ (show vals))
                                  return (Just dummyAddr)
catchExternal "L_print" es = do
                                  vals <- mapM eval es 
	                          msg (unlines (Prelude.map show vals))
                                  return (Just dummyAddr)
catchExternal _ es = return Nothing

-- allocate an object of given size
halloc_obj :: Integer -> Intp Val
halloc_obj sz = do
	          hp@(A hpa) <- getReg hpReg
                  topts <- getTraceOpts 
                  let traceme = AllocTrace `elem` topts
                  when traceme $
	            trace (".. Allocating object at address " ++ (show (A hpa)) ++ "\tsize: " ++ (show sz)) (return ())
	          let new_hp = hp+(I sz)
	          when (new_hp>hpEnd) $
                   error "PANIC: heap overflow"
	          setReg hpReg new_hp
	          return hp

init_obj :: Address -> Integer -> Intp ()
init_obj a sz = do
	              --(A a) <- obj
	              mm <- getM
	              setM (Map.union (Map.fromList (Prelude.map (\ x -> (x,(A 0))) [a..a+sz])) mm)
	              return ()

-- allocate an array of given size
halloc_arr :: Integer -> Intp Val
halloc_arr sz = do
	          hp@(A hpa) <- getReg hpReg
                  topts <- getTraceOpts 
                  let traceme = AllocTrace `elem` topts
                  when traceme $
	            trace (".. Allocating array at address " ++ (show (A hpa)) ++ "\tsize: " ++ (show sz)) (return ())
	          let new_hp = hp+(I (sz+1))
	          when (new_hp>hpEnd) $
                   error "PANIC: heap overflow"
	          setReg hpReg new_hp
	          return (hp+(I 1))  -- NB: length at index: -1

init_arr :: Address -> Integer -> Intp ()
init_arr a sz = do
	              --(A a) <- obj
	              mm <- getM
	              setM (Map.union (Map.fromList ((a-1,(I sz)):(Prelude.map (\ x -> (x,(I 99))) [a..a+sz]))) mm)
	              return ()

-- DUMMY
addrOfFrame _ = 0

mypow m 0 = 1
mypow m n = m*(mypow m (n-1))

myxor x y = x/=y

valToBool 0 = False
valToBool _ = True

boolToVal False = 0
boolToVal True = 1